@@ -12,8 +12,19 @@ my $pubnub = PubNub::PubSub->new(
     channel => $ENV{PUBNUB_CHANNEL} || 'sandbox',
 );
 
-my $history = $pubnub->history(100);
-
-print Dumper(\$history);
+my $history = $pubnub->history({
+    count => 5,
+    reverse => "true",
+});
+while (1) {
+    print Dumper(\$history);
+    last unless @{$history->[0]}; # no messages
+    sleep 1;
+    $history = $pubnub->history({
+        count => 5,
+        reverse => "true",
+        start => $history->[2]
+    });
+}
 
 1;
\ No newline at end of file
