@@ -1,8 +1,8 @@
 package PubNub::PubSub;
 
 use strict;
+use warnings;
 use v5.10;
-our $VERSION = '1.0.0';
 
 use Carp;
 use Mojo::JSON qw/encode_json/;
@@ -11,13 +11,15 @@ use Mojo::Util qw/url_escape/;
 
 use PubNub::PubSub::Message;
 
-sub new {
-    my $class = shift;
-    my %args  = @_ % 2 ? %{$_[0]} : @_;
+our $VERSION = '1.0.1';
 
-    $args{host} ||= 'pubsub.pubnub.com';
-    $args{port} ||= 80;
-    $args{timeout} ||= 60; # for ua timeout
+sub new {    ## no critic (RequireArgUnpacking)
+    my $class = shift;
+    my %args = @_ % 2 ? %{$_[0]} : @_;
+
+    $args{host}          ||= 'pubsub.pubnub.com';
+    $args{port}          ||= 80;
+    $args{timeout}       ||= 60;                    # for ua timeout
     $args{publish_queue} ||= [];
 
     my $proto = ($args{port} == 443) ? 'https://' : 'http://';
@@ -34,7 +36,7 @@ sub __ua {
     my $ua = Mojo::UserAgent->new;
     $ua->max_redirects(3);
     $ua->inactivity_timeout($self->{timeout});
-    $ua->proxy->detect; # env proxy
+    $ua->proxy->detect;    # env proxy
     $ua->cookie_jar(0);
     $ua->max_connections(100);
     $self->{ua} = $ua;
@@ -42,7 +44,7 @@ sub __ua {
     return $ua;
 }
 
-sub publish {
+sub publish {              ## no critic (RequireArgUnpacking)
     my $self = shift;
 
     my %params = @_ % 2 ? %{$_[0]} : @_;
@@ -51,19 +53,20 @@ sub publish {
     my $ua = $self->__ua;
 
     my @steps = map {
-             my $ref = $_;
-             my $url = $ref->{url};
-             sub {
-                 my $delay = shift;
-                 my $end = $delay->begin;
-                 $ua->get($url => sub {
+        my $ref = $_;
+        my $url = $ref->{url};
+        sub {
+            my $delay = shift;
+            my $end   = $delay->begin;
+            $ua->get(
+                $url => sub {
                     $callback->($_[1]->res, $ref->{message}) if $callback;
                     $end->();
-                  });
-             }
+                });
+            }
     } $self->__construct_publish_urls(%params);
 
-    Mojo::IOLoop->delay(@steps)->wait;
+    return Mojo::IOLoop->delay(@steps)->wait;
 }
 
 sub __construct_publish_urls {
@@ -79,13 +82,18 @@ sub __construct_publish_urls {
 
     return map {
         my $json = $_->json;
-        my $uri = Mojo::URL->new( $self->{web_host} . qq~/publish/$pub_key/$sub_key/0/$channel/0/~ . url_escape($json) );
+        my $uri  = Mojo::URL->new($self->{web_host} . qq~/publish/$pub_key/$sub_key/0/$channel/0/~ . url_escape($json));
         $uri->query($_->query_params(\%params));
-        { url => $uri->to_string, message => $_ };
-    } map { PubNub::PubSub::Message->new($_) } @{$params{messages}};
+        {
+            url     => $uri->to_string,
+            message => $_
+        };
+        } map {
+        PubNub::PubSub::Message->new($_)
+        } @{$params{messages}};
 }
 
-sub subscribe {
+sub subscribe {    ## no critic (RequireArgUnpacking)
     my $self = shift;
     my %params = @_ % 2 ? %{$_[0]} : @_;
 
@@ -102,18 +110,75 @@ sub subscribe {
     my $tx = $ua->get($self->{web_host} . "/subscribe/$sub_key/$channel/0/$timetoken");
     unless ($tx->success) {
         # for example $tx->error->{message} =~ /Inactivity timeout/
-        return $self->subscribe(%params, timetoken => $timetoken);
+
+        # This is not a traditional goto. Instead it exits this function
+        # and re-enters with @ as params.
+        #
+        # see goto docs, this is basically a method call which exits the current
+        # function first.  So no extra call stack depth.
+        sleep 1;
+        @_ = ($self, %params, timetoken => $timetoken);
+        goto &subscribe;
     }
     my $json = $tx->res->json;
+    my @cb_args = $params{raw_msg} ? ($json) : (@{$json->[0]});
 
-    my $rtn = $callback ? $callback->(@{ $json->[0] }) : 1;
+    my $rtn = $callback ? $callback->(@cb_args) : 1;
     return unless $rtn;
 
     $timetoken = $json->[1];
     return $self->subscribe(%params, timetoken => $timetoken);
 }
 
-sub history {
+sub subscribe_multi {    ## no critic (RequireArgUnpacking)
+    my $self = shift;
+    my %params = @_ % 2 ? %{$_[0]} : @_;
+    croak 'channels must be an arrayref'
+        unless ref($params{channels}) =~ /ARRAY/;
+    croak 'callback must be a hashref or coderef'
+        unless ref($params{callback}) =~ /(HASH|CODE)/;
+
+    my $callback;
+    if (ref($params{callback}) =~ /HASH/) {
+        for (keys %{$params{callback}}) {
+            croak "Non-coderef value found for callback key $_"
+                unless ref($params{callback}->{$_}) =~ /CODE/;
+        }
+        $callback = sub {
+            my ($obj) = @_;
+            my (undef, undef, $channel) = @$obj;
+            my $cb_dispatch = $params{callback};
+            unless ($channel) {    # on connect messages
+                goto $cb_dispatch->{on_connect}
+                    if exists $cb_dispatch->{on_connect};
+                return 1;
+            }
+            if (exists $cb_dispatch->{$channel}) {
+
+                # these are verified coderefs, so replacing the current stack
+                # frame with a call to the function.  They will *not* jump to
+                # a label or other points.  Basically this just lets us pretend
+                # that this was called directly by subscribe above.
+                goto $cb_dispatch->{$channel};
+            } elsif (exists $cb_dispatch->{'_default'}) {
+                goto $cb_dispatch->{_default};
+            } else {
+                warn 'Using callback dispatch table, cannot find channel callback' . ' and _default callback not specified';
+                return;
+            }
+        };
+    }
+    $callback = $params{callback} unless ref $callback;
+
+    my $channel_string = join ',', @{$params{channels}};
+    return $self->subscribe(
+        channel  => $channel_string,
+        callback => $callback,
+        raw_msg  => 1
+    );
+}
+
+sub history {    ## no critic (RequireArgUnpacking)
     my $self = shift;
 
     if (scalar(@_) == 1 and ref($_[0]) ne 'HASH' and $_[0] =~ /^\d+$/) {
@@ -147,6 +212,8 @@ PubNub::PubSub - Perl library for rapid publishing of messages on PubNub.com
 =head1 SYNOPSIS
 
     use PubNub::PubSub;
+    use 5.010;
+    use Data::Dumper;
 
     my $pubnub = PubNub::PubSub->new(
         pub_key => 'demo', # only required for publish
@@ -230,6 +297,36 @@ set ENV MOJO_USERAGENT_DEBUG to debug
 
 subscribe channel to listen for the messages.
 
+Arguments are:
+
+=over
+
+=item callback
+
+Callback to run on the channel
+
+=item channel
+
+Channel to listen on, defaults to the base object's channel attribute.
+
+=item subkey
+
+Subscription key.  Defaults to base object's subkey attribute.
+
+=item raw_msg
+
+Pass the whole message in, as opposed to the json element of the payload.
+
+This is useful when you need to process time tokens or channel names.
+
+The format is a triple of (\@messages, $timetoken, $channel).
+
+=item timetoken
+
+Time token for initial request.  Defaults to 0.
+
+=back
+
     $pubnub->subscribe({
         callback => sub {
             my (@messages) = @_;
@@ -242,6 +339,29 @@ subscribe channel to listen for the messages.
 
 return 0 to stop
 
+=head2 subscribe_multi
+
+Subscribe to multiple channels.  Arguments are:
+
+=over
+
+=item channels
+
+an arrayref of channel names
+
+=item callback
+
+A callback, either a coderef which handles all requests, or a hashref dispatch
+table with one entry per channel.
+
+If a dispatch table is used a _default entry catches all unrecognized channels.
+If an unrecognized channel is found, a warning is generated and the loop exits.
+
+The message results are passed into the functions in raw_msg form (i.e. a tuple
+ref of (\@messages, $timetoken, $channel) for performance reasons.
+
+=back
+
 =head2 publish
 
 publish messages to channel
@@ -379,8 +499,8 @@ for example, to fetch all the rows in history
 
 This module effectively runs a Mojolicious application in the background.  For
 those parts of JSON which do not have a hard Perl equivalent, such as booleans,
-the Mojo::JSON module's semantics work.  This means that JSON bools are 
-handled as references to scalar values 0 and 1 (i.e. \0 for false and \1 for 
+the Mojo::JSON module's semantics work.  This means that JSON bools are
+handled as references to scalar values 0 and 1 (i.e. \0 for false and \1 for
 true).
 
 This has changed since 0.08, where True and False were used.
@@ -393,44 +513,4 @@ L<https://github.com/binary-com/perl-pubnub-pubsub>
 
 Binary.com E<lt>fayland@gmail.comE<gt>
 
-=head1 LICENSE AND COPYRIGHT
-
-Copyright 2014- binary.com.
-
-This program is free software; you can redistribute it and/or modify it
-under the terms of the the Artistic License (2.0). You may obtain a
-copy of the full license at:
-
-L<http://www.perlfoundation.org/artistic_license_2_0>
-
-Any use, modification, and distribution of the Standard or Modified
-Versions is governed by this Artistic License. By using, modifying or
-distributing the Package, you accept this license. Do not use, modify,
-or distribute the Package, if you do not accept this license.
-
-If your Modified Version has been derived from a Modified Version made
-by someone other than you, you are nevertheless required to ensure that
-your Modified Version complies with the requirements of this license.
-
-This license does not grant you the right to use any trademark, service
-mark, tradename, or logo of the Copyright Holder.
-
-This license includes the non-exclusive, worldwide, free-of-charge
-patent license to make, have made, use, offer to sell, sell, import and
-otherwise transfer the Package with respect to any patent claims
-licensable by the Copyright Holder that are necessarily infringed by the
-Package. If you institute patent litigation (including a cross-claim or
-counterclaim) against any party alleging that the Package constitutes
-direct or contributory patent infringement, then this Artistic License
-to you shall terminate on the date that such litigation is filed.
-
-Disclaimer of Warranty: THE PACKAGE IS PROVIDED BY THE COPYRIGHT HOLDER
-AND CONTRIBUTORS "AS IS' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
-THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
-PURPOSE, OR NON-INFRINGEMENT ARE DISCLAIMED TO THE EXTENT PERMITTED BY
-YOUR LOCAL LAW. UNLESS REQUIRED BY LAW, NO COPYRIGHT HOLDER OR
-CONTRIBUTOR WILL BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, OR
-CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THE PACKAGE,
-EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
 =cut
